#!/usr/bin/perl
#this program takes a C-source with MS-VC-like assembler (with __asm)
#and outputs a source compilable by gcc (normally...)
#you must compile the produced source with the option "-masm=intel"
#(the version of gcc must be at least 3.1)


#with samenumline the lines of the input and output correspond.
#it is useful to debug a program
$samenumline=1;
#use locallabels or macro if you want to compile with the -O3 option
$macro=0;
$locallabels=0;
$PARANO=0;

if($ARGV[0] =~ /^(-h|--help|-help)$/) {
    print STDERR "usage: msof2gcc.pl [-locallabels|-macro] [ prog.c ] > prog2.c\n";
    exit 0;
}



if($ARGV[0] eq "-macro") {
    $macro = 1;
    shift @ARGV;
} elsif ($ARGV[0] eq "-locallabels") {
    $locallabels = 1;
    shift @ARGV;
}

$registers="(ax|eax|ebx|ecx|edi|edx|esi|mm0|mm1|mm2|mm3|mm4|mm5|mm6|mm7".
    "|xmm0|xmm1|xmm2|xmm3|xmm4|xmm5|xmm6|xmm7)"; 

$numbloc=0;

sub init_bloc {
    $numbloc++;
    %identifiers_bloc=();
    %registers_bloc=();
    if($locallabels) {
	%labels=();
	$nlabel=0;
    }
}

sub end_bloc {
    my $first=1;
    if($macro) {
	print "\".endm\\n\\t\"  \"BB${numbloc}\\n\\t\" ";
    }
    print ": :";
    if(not $samenumline) {print "\n";}
    foreach(keys %identifiers_bloc) {
	if(not $first) {print ",";} else{$first=0}
	print "[$identifiers_bloc{$_}] \"m\" ($_)";
    }
    if(not $samenumline) {print "\n";}
    print ":";
    if(not $samenumline) {print "\n";}
    foreach(keys %registers_bloc) {
	print "\"%$_\",";
    }
    print "\"memory\"";
    if(not $samenumline) {print "\n";}
}

sub conv_identifier {
    my ($arg);
    $arg=$_[0];
    if(not defined $identifiers_bloc{$arg}) {
	$_=$arg;
	s/\./_/g;
	$identifiers_bloc{$arg}=$_;
    }
    #print STDERR "$arg\n";
    return "%[$identifiers_bloc{$arg}]";
}

sub conv_register {
    my ($arg);
    $arg=$_[0];
    $registers_bloc{$arg}=1;
    return "%%$arg";
}

if($locallabels) {
    sub conv_label {
	my ($arg, $label);
	$arg=$_[0];
	$label=$_[1];
	if(not defined $labels{$arg}) {
	    $labels{$arg}="${nlabel}f";
	    $nlabel++;
	}
	if($label) {
	    $_=$labels{$arg};
	    s/f$//;
	    $labels{$arg}="${_}b";
	    return $_;
	}
	return "$labels{$arg}";
    }
}

sub convert_formula {
    my ($arg,$ret,$reg,$reg2);
    $arg=$_[0];
    #print STDERR "formula=$arg\n";
    $_=$arg;
    if (m/^[a-zA-Z_][a-zA-Z_0-9]*$/) {
	if(m/^$registers$/i) {
	    $ret=conv_register($arg);
	} else {
	    $ret=conv_identifier($arg);
	}
    } else {
	$_=$arg;
	#print STDERR "DEBUT BOUCLE\n";
	while(m/^(.*[^%a-zA-Z])??($registers)([^a-zA-Z0-9].*)??$/) {
	    $reg=$2;
	    $reg2=conv_register($reg);
	    #print STDERR "$reg --> $reg2\n";
	    s/^((.*[^%a-zA-Z])??)$reg(([^a-zA-Z0-9].*)??)$/\1$reg2\3/;
	}
	#print STDERR "FIN BOUCLE\n";
	$ret=$_;
    }
    return $ret;
}

sub convert_arg {
    my ($arg,$arg2,$pref,$suff);
    $arg=$_[0];
    #print STDERR "__________ $arg _______________\n";
    if($arg eq "") { return "";}
    $_=$arg;

    if(m/^(dword|byte|qword)[\s]+ptr[\s]*\[([^\]]*)\]$/i) {
	$pref=$1;
	$suff=$2;
	if($suff =~ m/^[a-zA-Z_][a-zA-Z_0-9]*$/ and
	   not $suff =~ m/^$registers$/i) {
	    #this generates dword ptr[dword ptr[...]]
	    #it compiles but is it correct ?
	    #$arg2="$pref ptr[".conv_identifier($suff)."]";
	    $arg2=conv_identifier($suff);
	    $NOTSURE=1;
	} else {
	    $suff=convert_formula($suff);
	    if(not "$cmd" =~ /^(movq|movd|punpckldq)$/) {
		$arg2="$pref ptr[$suff]";
	    } else {
		$arg2="[$suff]"; # est-ce correct ??
		$NOTSURE=1;
	    }
	}
    } elsif (m/^\[([^\]]*)\]$/) {
	$arg2="[".convert_formula($1)."]";
    } elsif (m/^[a-zA-Z_][a-zA-Z_0-9]$*/) {
	if(m/^$registers$/i) {
	    $arg2=conv_register($arg);
	} else {
	    $arg2=conv_identifier($arg);
	}
    } elsif (m/^[0-9]*$/) {
	$arg2=$arg;
    } else {
	print STDERR "BIG PB with __${arg}__";
	exit(1);
    }
    #print STDERR "'${arg}' --> '${arg2}'\n";
    #print STDERR "${arg}\n"; # to generate the registers !!
    return $arg2;
}


sub convert {
    my ($micros,$args,$arg1,$arg2); #MANDATORY parenthesis
    $NOTSURE=0;
    $micros=$_[0];
    #print STDERR "inst=\@\@${micros}\@\@\n";
    $_=$micros;
    if(!m/([a-zA-Z_][a-zA-Z_0-9]*:??)([\s]+(([^\s].*)??))??$/) {
	print STDERR "BIG PB avec __${micros}__\n";
	exit(1);
    }
    $cmd=$1;
    $args=$3;
    
    #Some command alter registers which are not in the argument
    #In fact I only notic the command "cpuid", but if
    #there others, add them here !!!
    if($cmd eq "cpuid") {
	foreach("eax", "ebx","ecx","edx","esi","edi") {
	    conv_register($_);
	}
    }

    if($cmd =~ /:$/) {
	if($macro) {
	    $_=$cmd; s/:$//; $cmd=$_;
	    $gcc="B$numbloc.$cmd.\\\\\@: $args";
	} elsif($locallabels) {
	    my ($label);
	    $_=$cmd; s/:$//; $cmd=$_;
	    $label=conv_label($cmd,1);
	    $gcc="$label: $args"
	} else {
	    $gcc="B$numbloc.$cmd $args";
	}
    } elsif($cmd eq "align") {
	$gcc=".align $args";
    } elsif($args eq "") {
	$gcc="$cmd $args";
    } else {
	$_=$args;
	if(! m/^([^,]*[^\s])([\s]*,[\s]*([^\s][^,]*))??$/) {
	    print STDERR "BIG PB avec __${args}__\n";
	    exit(1);
	}
	$arg1=$1;
	$arg2=$3;
	
	#print STDERR "arg1=$arg1 arg2=$arg2\n";
	#faudra peut-etre traiter les labels aussi....
	if(not $cmd =~ m/^(jmp|jnz|jg|jl|jne)$/i) {
	    $arg1=convert_arg($arg1);
	    $arg2=convert_arg($arg2);
	} else {
	    if($macro) {
		$arg1="B$numbloc.$arg1.\\\\\@";
	    } elsif($locallabels) {
		$arg1=conv_label($arg1,0);
	    } else {
		$arg1="B$numbloc.$arg1";
	    }
	}
	if($arg2 ne "") {
	    $args="$arg1, $arg2";
	} else {
	    $args=$arg1;
	}
	$gcc ="$cmd $args";
    }
    
    if($NOTSURE and $PARANO) {
	print STDERR "'${micros}' --> '${gcc}' NOT SURE\n";
	if(not $samenumline) {
	    $gcc = "$gcc \\n\\t\"\n\"\#\# NOT SURE originally ${micros}";
	}
    }
    #print STDERR "'${micros}' --> '${gcc}'\n";
    return $gcc;
}

while(<>) {
    if(/__asm/) {
	if(m/__asm[\s].*;/) {
	    s/__asm[\s]+(.*)[\s]*;/__asm__ __volatile__\(\"\1\\n\\t\"\);/;
	    print;
	    next;
	}
	s/__asm/__asm__ __volatile__/;
	do {
	    print;
	    $_=<>; # must be in scalar context
	} while(! m/\{/);
	init_bloc;
	if(!$macro) {
	    s/\{/\(/;
	} else {
	    s/\{/\( \"\.macro BB$numbloc\\n\\t\" /;
	}
	print;
	$_=<>;
	do {
	    s/^([\s])*([a-zA-Z])/\1\"\2/;
	    if(!m&^(.*)\"([^/]+(/[^/]+)*[^\s])([\s]*(//.*)??)$&) {
		print;
	    } else {
		$avantassemb=$1;
		$assemb=$2;
		$apresassemb=$4;
		chomp($assemb);
		$assemb2=convert($assemb);
		chomp($apresassemb);
		print "$avantassemb\"$assemb2\\n\\t\"$apresassemb\n";
	    }
	    $_=<>; # must be in scalar context
	} while(! m/\}/);
	end_bloc;
	s/\}/\);/;
	print;
    }
    else {
	s/([^_])_asm[\s]*(.*);/\1__asm__ __volatile\(\"\2\\n\\t\"\);/;
	print;
    }
}

