/*********************************************************************************
 *                                                                               *
 * kizkool32.c                                                                   *
 *                                                                               *
 * This code has been developped by KizKoool and others for Multidec-SE          *
 * and adapted by Keuleu for xawdecode                                           *
 *                                                                               *
 *                                                                               *
 * This program is free software; you can redistribute it and/or                 *
 * modify it under the terms of the GNU General Public License                   *
 * as published by the Free Software Foundation; either version 2                *
 * of the License, or (at your option) any later version.                        *
 *                                                                               *
 *                                                                               *
 * This program is distributed in the hope that it will be useful,               *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of                *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                 *
 * GNU General Public License for more details.                                  *
 *                                                                               *
 *                                                                               *
 * You should have received a copy of the GNU General Public License             *
 * along with this program; if not, write to the Free Software                   *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.    *
 * Or, point your browser to http://www.gnu.org/copyleft/gpl.html                *
 *                                                                               *
 *                                                                               *
 *********************************************************************************/
#include <stdio.h>
#include "config.h"
#define KIZKOOOL_32
#include "filtre.h"
#include "kizkool.h"
void set_bluefilter_params (int filter, ...);

#ifdef ENABLE_KIZKOOL32

#define DWORD long

// KizKoool extensions
#define ZEILENZAHL           287

/*
int nUThreshold;
int nVThreshold;
int nYUVThreshold;
*/
static int nPixThreshold;
static int iLine;

// Defined in kizkool.c
//extern int iUFilterThresholdPCent;
//extern int iVFilterThresholdPCent;
//extern int iYUVFilterThresholdPCent;
//extern int iPixFilterThreshold;

extern int debug;

extern KIZKOOL_PARAM kizkool_param[NB_KIZKOOL_FILTER];

extern int TypeImage;
extern unsigned int CurrentFrame;
extern int CurrentX;
extern int MovieScale;

void
init_kizkool32_filter (void)
{
  if (debug)
    fprintf(stderr, "kizkool32: init filter\n");
  set_bluefilter_params(BLUEFILTER_K32, KIZKOOL_SMUDGE, SMUDGEFILTER_OFF, 0);
  set_bluefilter_params(BLUEFILTER_K32, KIZKOOL_UTHRESHOLD, (U_FILTER_THRESHOLD_K32*65535)/100, 0);
  set_bluefilter_params(BLUEFILTER_K32, KIZKOOL_VTHRESHOLD, (V_FILTER_THRESHOLD_K32*65535)/100, 0);
  set_bluefilter_params(BLUEFILTER_K32, KIZKOOL_YUVTHRESHOLD, (YUV_FILTER_THRESHOLD_K32*65535)/100, 0);
  set_bluefilter_params(BLUEFILTER_K32, KIZKOOL_PIXFILTERTHRESHOLD, (PIX_FILTER_THRESHOLD_K32*65535)/100, 0);
  kizkool_param[BLUEFILTER_K32 - 1].nUThreshold =
    (int) (U_FILTER_THRESHOLD_K32 * 0.01 * 0xff);
  kizkool_param[BLUEFILTER_K32 - 1].nVThreshold =
    (int) (V_FILTER_THRESHOLD_K32 * 0.01 * 0xff);
  kizkool_param[BLUEFILTER_K32 - 1].nYUVThreshold =
    (int) (YUV_FILTER_THRESHOLD_K32 * 0.01 * 0xff);

}

//
// Filtres KizKoool SECAM v3.2
//
// Filtre lignes violettes + filtre anti-bavures
//
// Premires fonctions de traitement SECAM par Fantasio
// Ancien filtre bleu ASM par 1konnu
//
// Nouvel algorithme de dtection des lignes satures + bugfixes +
// traduction en C du filtre bleu et optimisations par KizKoool
//
// Ide et code original du filtre anti-bavures par 1Konnu
//
// Adaptation  MultiDec + optimisations +
// rephasage dynamique du filtre anti-bavures par KizKoool
//
// Ide de l'exploitation de la variable CurrentFrame par E-nek
// Symbiose optimale des 2 filtres par Quenotte et KizKoool
//
void
SecamFiltering32 (DWORD ** pLinesAddress, unsigned short *pSortLines)
{
  // Taille maximale d'une frame 768*287
  static BYTE ColorBackup[768 * ZEILENZAHL];
  BYTE *pColorQuelle, *pColorBuffer, *pQuelle, *pSource, *pix_courant;

  nPixThreshold =
    (int) ((CurrentX * kizkool_param[BLUEFILTER_K32 - 1].iPixFilterThreshold)
           / 200);

  for (iLine = 1 + MovieScale; iLine < ZEILENZAHL - MovieScale; iLine++)
    {
      // pQuelle: pointeur sur la ligne source dans l'image dcode:
      pQuelle = (BYTE *) pLinesAddress[pSortLines[iLine]];
      // pColorBuffer: pointeur sur les couleurs non satures du buffer
      pColorBuffer = (BYTE *) (ColorBackup + CurrentX * iLine);

      //Test de saturation de la ligne

      if (LineSaturation (pLinesAddress[pSortLines[iLine]], CurrentX >> 1))
        {
          // Ligne sature

          // Restauration de la saturation  partir du buffer de lignes non satures

          for (pix_courant = pQuelle + 1;
               pix_courant < pQuelle + (CurrentX << 1);
               pix_courant = pix_courant + 4)
            {
              *pix_courant = *pColorBuffer;
              *(pix_courant + 2) = *(pColorBuffer + 1);
              pColorBuffer = pColorBuffer + 2;
            }
        }
      else
        {
          // Ligne non sature

          // 1) Application du filtre anti-bavures

          if (iLine != 1 + MovieScale)
            {
              // pColorQuelle: pointeur sur la ligne precedente dans l'image dcode:
              pColorQuelle = (BYTE *) pLinesAddress[pSortLines[iLine - 1]];

              if ((((iLine % 2) == 0) && TypeImage)
                  || (((iLine % 2) != 0) && !TypeImage))
                {
                  // ligne paire de l'image de type 1 ou ligne impaire de l'image de type 2
                  pSource = pQuelle + 3;
                  for (pix_courant = pColorQuelle + 3;
                       pix_courant < pColorQuelle + (CurrentX << 1);
                       pix_courant = pix_courant + 4)
                    {
                      *pSource = *pix_courant;
                      pSource = pSource + 4;
                    }
                }
              else
                {
                  // ligne impaire de l'image de type 1 ou ligne paire de l'image de type 2
                  pSource = pQuelle + 1;
                  for (pix_courant = pColorQuelle + 1;
                       pix_courant < pColorQuelle + (CurrentX << 1);
                       pix_courant = pix_courant + 4)
                    {
                      *pSource = *pix_courant;
                      pSource = pSource + 4;
                    }
                }
            }

          // 2) Stockage de la ligne dans le buffer de lignes non satures

          for (pix_courant = pQuelle + 1;
               pix_courant < pQuelle + (CurrentX << 1);
               pix_courant = pix_courant + 4)
            {
              *pColorBuffer = *pix_courant;
              *(pColorBuffer + 1) = *(pix_courant + 2);
              pColorBuffer = pColorBuffer + 2;
            }
        }
    }                            // boucle for
}

//
// Test de saturation de ligne par KizKoool
//
// Paramtres:
// pQuelle: Pointeur vers la ligne  tester
// CountX: Nombre de pixels  tester
//
static int
LineSaturation (DWORD * pQuelle, int CountX)
{
  DWORD *pix_courant;
  long U, V, Y, Seuil_NbPix;
  int nUThreshold = kizkool_param[BLUEFILTER_K32 - 1].nUThreshold;
  int nVThreshold = kizkool_param[BLUEFILTER_K32 - 1].nVThreshold;
  int nYUVThreshold = kizkool_param[BLUEFILTER_K32 - 1].nYUVThreshold;

  Seuil_NbPix = nPixThreshold;
  pix_courant = pQuelle;

  while ((pix_courant < pQuelle + CountX) && (Seuil_NbPix > 0))
    {
      U = *pix_courant >> 8 & 0xff;
      V = *pix_courant >> 24 & 0xff;
      Y = ((*pix_courant >> 16 & 0xff) + (*pix_courant & 0xff)) >> 2;

      if (!
          ((((U >> 1) + V) > Y + nYUVThreshold)
           && ((U > nUThreshold) || (V > nVThreshold))))
        {
          Seuil_NbPix -= 1;
        }
      pix_courant++;
    }

  return (Seuil_NbPix > 0);
}


//
// Rephasage dynamique du filtre anti-bavures par KizKoool
// Algorithme bas sur les proprits des lignes satures
//
void
FindFrameType32 (DWORD ** pLinesAddress, unsigned short *pSortLines)
{
  DWORD *pQuelle, *pColorQuelle, *pLineBackup;
  static DWORD LineColorBackup[768];
  unsigned long MasqueToutSaufU, MasqueToutSaufV;
  int nb_saturees1, nb_saturees2;
  nb_saturees1 = 0;
  nb_saturees2 = 0;

  MasqueToutSaufU = 0xFFFF00FF;    // Garde tout sauf la composante U
  MasqueToutSaufV = 0x00FFFFFF;    // Garde tout sauf la composante V

  pLineBackup = (DWORD *) LineColorBackup;

  nPixThreshold =
    (int) ((CurrentX * kizkool_param[BLUEFILTER_K32 - 1].iPixFilterThreshold)
           / 200);

  // Dtection des frames loupes

  if (NbFramesManquantes () % 2 == 1)
    TypeImage = !(TypeImage);

  // Dtection du type d'image

  for (iLine = ZEILENZAHL - 64; iLine < ZEILENZAHL - 32; iLine++)
    {
      pQuelle = (DWORD *) pLinesAddress[pSortLines[iLine]];
      pColorQuelle = (DWORD *) pLinesAddress[pSortLines[iLine - 1]];

      if ((((iLine % 2) == 0) && TypeImage)
          || (((iLine % 2) != 0) && !TypeImage))
        {
          // ligne paire de l'image de type 1
          // ou ligne impaire de l'image de type 2
          nb_saturees1 +=
            TestCorrectionUV (pQuelle, pLineBackup, pColorQuelle,
                              MasqueToutSaufV);
          nb_saturees2 +=
            TestCorrectionUV (pQuelle, pLineBackup, pColorQuelle,
                              MasqueToutSaufU);
        }
      else
        {
          // ligne impaire de l'image de type 1
          // ou ligne paire de l'image de type 2
          nb_saturees1 +=
            TestCorrectionUV (pQuelle, pLineBackup, pColorQuelle,
                              MasqueToutSaufU);
          nb_saturees2 +=
            TestCorrectionUV (pQuelle, pLineBackup, pColorQuelle,
                              MasqueToutSaufV);
        }
    }

  // Test de rephasage

  if ((nb_saturees1 > nb_saturees2 + 4)
      && (min (nb_saturees1, nb_saturees2) < 4))
    TypeImage = !(TypeImage);
}

//
// Dtection des frames manquantes par KizKoool
// Sur une ide d'exploiter la variable CurrentFrame d'E-nek
//
// Valeur de retour:
// Nombre de frames non acquises
//
/*
static int
NbFramesManquantes (void)
{
  static int PrecedentFrame = -1;
  int NbFramesLoupees;

  if (CurrentFrame > PrecedentFrame)
    NbFramesLoupees = CurrentFrame - PrecedentFrame - 1;
  else
    NbFramesLoupees = CurrentFrame - PrecedentFrame + 4;

  PrecedentFrame = CurrentFrame;

  return NbFramesLoupees;
}
*/
static int
NbFramesManquantes(void)
{
  static unsigned int PrecedentFrame=0;
  unsigned int NbFramesLoupees;

  NbFramesLoupees = (CurrentFrame - PrecedentFrame)/2 -1;
  PrecedentFrame = CurrentFrame;
  //printf("NbFramesLoupees  = %u\n", NbFramesLoupees);

  return (int)NbFramesLoupees;
}

//
// Test de correction des Couleurs UV par KizKoool
//
// Paramtres :
// pQuelle      : Source
// pSauve       : Destination 
// pColorQuelle : Rfrence pour reconstituer la couleur
// MasqueUV     : Masque de correction
// 
static int
TestCorrectionUV (DWORD * pQuelle, DWORD * pSauve, DWORD * pColorQuelle,
                  DWORD MasqueUV)
{
  unsigned long MasqueInverse;
  DWORD *pix_courant;
  MasqueInverse = ~MasqueUV;

  for (pix_courant = pSauve; pix_courant < pSauve + (CurrentX >> 1);
       pix_courant++)
    {
      *pix_courant =
        ((*pQuelle & MasqueUV) | (*pColorQuelle & MasqueInverse));
      pColorQuelle++;
      pQuelle++;
    }
  return LineSaturation (pSauve, CurrentX >> 1);
}
#endif
